-- LibreGates/Gates_lib.vhdl
-- created 20190618
-- version 20190809 (restart)
-- version sam. août 10 08:06:58 CEST 2019
-- version dim. août 11 17:36:51 CEST 2019
-- version lun. août 12 18:36:04 CEST 2019
-- version mar. août 13 23:14:14 CEST 2019
-- version mar. nov. 19 07:04:27 CET 2019 : off-by-one in histogram display
-- version jeu. nov. 21 03:50:23 CET 2019 : changing from fast/trace to run_mode,
--    lookitupX_fast merged,  Gate_Census now with no parameter list
-- version dim. déc.  1 14:00:01 CET 2019 : preparing for 64 bits for histograms and changes
-- version ven. déc. 13 17:07:51 CET 2019 : adding 16-entries LUTs
-- version lun. mars 23 03:47:01 CET 2020 : v2.8 adds "simple" gate types (no LUT)
-- version mar. juil. 14 23:02:32 CEST 2020 : added function ComputeLUT16()
-- version sam. août 29 23:57:38 CEST 2020 : v2.10 with the new internal representation system
-- version lun. août 31 22:17:01 CEST 2020 : functions trim_name and trim_spaces moved to SLV_utils
--     avoiding naming collision with display_histogram => display_gates_histograms
--      moved all the PA3-specific code to PA3_gategen.vhdl which is now Gates_lib.vhdl
--
-- Released under the GNU AGPLv3 license
--
-- This file is the root library that is used by most other files
-- that support analysis/tracing/etc. such as 
-- PA3_gategen.vhdl, PA3_components.vhdl and PA3_definitions.vhdl.

Library ieee;
    use ieee.std_logic_1164.all;
use STD.textio.all;
Library SLV;
    use SLV.SLV_utils.all;
Library Histo;
    use Histo.Histograms.all;

package Gates_lib is

  type integer_array is array(integer range <>) of integer;
  type integer_array_ptr is access integer_array;
--  type integer_array_array is array(integer range <>) of integer_array_ptr;
--  type integer_array_array_ptr is access integer_array_array;

  subtype LookupType1  is std_logic_vector(0 to  1);
  subtype LookupType2  is std_logic_vector(0 to  3);
  subtype LookupType4  is std_logic_vector(0 to  7);
  subtype LookupType8  is std_logic_vector(0 to 15);
  subtype LookupType16 is std_logic_vector(0 to 31);

  type enum_run_modes is (
    fast, trace, probe
  );
  shared variable run_mode : enum_run_modes := fast;

  type enum_gate_kind is (
    Gate_boolean,
    Gate_register,
    Gate_backwards
  );

  type KindShorthandType is array(enum_gate_kind) of string(1 to 4);
  constant KindShorthands : KindShorthandType := ( "Bool", "Reg.", "Back");

  -- this is a basic description of an input : a gate number and the port's number.
  -- this can then be assembled in an array to define a "net" (when associated to a driver).
  type input_address_record is
    record
      gate_number  : integer;
      gate_port    : integer;
    end record;
  type input_address_array is array (integer range <>) of input_address_record;
  type input_address_access is access input_address_array;

  -- this one is more sophisticated : inputs can be chained to create a "net".
  -- This can be linearised later, after the scan, to condense/compact the representation.
  type input_record_type is
    record
      source     : integer; -- number of the gate that drives this input. Negative if driven by input port.
      depth      : integer; -- how many gates to traverse to reach the closest input.
      next_input : integer; -- next gate in the linked list
      next_port  : integer; -- eventual port number of the next gate in the linked list
    end record;
  type input_records_array is array (integer range <>) of input_record_type;
  type input_records_access is access input_records_array;

  type output_record_type is
    record
           height : integer; -- how many gates to traverse to reach the closest output.
           fanout : integer; -- how many gates are driven by this gate
       first_sink : integer; -- number of the first gate in the linked list of sinks
       first_port : integer; -- number of the port of the first gate in the linked list of sinks
      sinks_array : input_address_access;  -- linear version of the linked list
    end record;
  type output_record_array is array (integer range <>) of output_record_type;
  type output_record_access is access output_record_array;

  -- This is a part of the netlist (when the gate numbers are negative) and defines the ports.
  -- notice the swap : the input vector is seen as outputs by the following gates and vice versa.
  shared variable output_records : input_records_access; -- the array of descriptors of the output vector
  shared variable input_records : output_record_access; -- array of the descriptors of the input


  -- The representation of a gate (and also the storage area for the histograms)
  --  should be better coded ! (but works well so far)
  constant SPECIAL : natural := 32; -- the index of the counter of "non-binary" codes
  type type_Histo16 is array (0 to SPECIAL) of big_sint_t; -- can store a LUT16, the "shadow/meta" LUT16 plus the "special"s counter
  type Gate_Descriptor is
    record
      gate_kind  : enum_gate_kind; -- boolean, DFF or backwards ?
      index      : natural;      -- the index of the last lookup
      last       : natural;      -- (size of the LUT) the last valid index (can be 0)
      inputs     : integer;      -- O to 4 inputs
      histo      : type_Histo16; -- the histogram
      curOut,                    -- the last result of lookup
      prevOut    : std_logic;    -- the previous result of the lookup
      changes    : big_uint_t;   -- how many times the output changed
      gatename   : line;         -- full path & name of the gate
      depth_min  : integer;      -- how many gates to traverse to reach the closest input.
      depth_max  : integer;      -- how many gates to traverse to reach the farthest input. <0 if zombie.
      LUT        : std_logic_vector(0 to 15); -- cache of the gate's LUT
      out_record : output_record_type;
      in_record  : input_records_array(0 to 3);
    end record;
  type Gate_Desc_Array is array(integer range <>) of Gate_Descriptor;
  type Gate_Desc_Access is access Gate_Desc_Array;
  shared variable GateList : Gate_Desc_Access;   -- The rest of the netlist
  shared variable gate_instance_counter,
                  flipflop_instance_counter,
                  backwards_instance_counter : integer := 0;

  constant inputs_LUT: integer_array(0 to 16) := (
         0 =>  0,
         1 =>  1,
         3 =>  2,
         7 =>  3,
        15 =>  4,
    others => -1);

  -- the list of registered gates:
  -- (in case the general list of gates contains holes,
  -- such as GND/VCC (LUT0) gates that are not listed)
  shared variable indexPtr : integer_array_ptr;
  shared variable registered_instance_counter : integer := 0;

  shared variable gate_select_number : integer := 0;
  shared variable bit_select_number : integer := -1;
  shared variable cycle_counter : big_uint_t := 0;
  shared variable flip_LUT : integer := 0;

  -- manage the file of exclusions:
  file exclude_file : TEXT;
  shared variable xcl_filename, current_vector : line;
  shared variable xcl_gatenum, xcl_oldgatenum : integer := -1;
  shared variable verbose_i, line_number : integer := 0;

-- Used when probing:
  shared variable probe_phase : std_logic := 'U';
  type slv2int_array is array (std_logic) of integer;
  shared variable probe_numbers : slv2int_array;
     -- values :
     --  integer'low : empty slot
     --   > 0 : netlist gate
     --  <= 0 : input signal

  impure function Gate_Census return integer;
  procedure update_select_gate(gate_number, bit_number, flip : integer; verb, filename : string);
  impure function RegisterLUT(L: std_logic_vector; fullpath: string;
     gate_number, lastbit : integer; exclude: std_logic_vector) return std_logic_vector;
  impure function RegisterBack(fullpath: string; gate_number : integer) return integer;
  procedure update_gates_histograms;
  procedure display_gates_histograms;

  impure function register_source(gate_number, input_number : integer; val : std_logic) return input_record_type;
  procedure  register_gate_source(gate_number, input_number : integer; val : std_logic);
  impure function probe_return(gate_number : integer) return std_logic;
  impure function GoBackwards (A        : std_logic;                     gate_number : integer) return std_logic;
  impure function LookItUp1 (A          : std_logic; LUT2 : LookupType2; gate_number : integer) return std_logic;
  impure function LookItUp2 (A, B       : std_logic; LUT4 : LookupType4; gate_number : integer) return std_logic;
  impure function LookItUp3 (A, B, C    : std_logic; LUT8 : LookupType8; gate_number : integer) return std_logic;
  impure function LookItUp4 (A, B, C, D : std_logic; LUT16: LookupType16;gate_number : integer) return std_logic;

  procedure get_another_vector;
  procedure init_get_vectors;

end Gates_lib;

-----------------------------------------------------------------------------------------

package body Gates_lib is

  -- just count the number of registered gates
  impure function Gate_Census return integer is
  begin
    gate_instance_counter := gate_instance_counter + 1;
    return gate_instance_counter;
  end Gate_Census;


  -- now that we know how many gates we have,
  -- allocate and initialise some shared variables with the generics,
  --   as well as the netlist/histograms
  -- Called by the component register_generics
  procedure update_select_gate(
    gate_number, bit_number, flip : integer;
    verb, filename : string) is
  begin
    report integer'image(gate_instance_counter) & "x A3P gates found.";
    if gate_number < 1 then
      report " gate number: " & integer'image(gate_number)
           & "  bit number: " & integer'image( bit_number);
      report " -- no gate function is altered from the generics --";
      report "List of the registered gates:";
    end if;
    -- register the generics and other parameters into the shared variables :
    gate_select_number := gate_number;
     bit_select_number :=  bit_number;
    if flip = 2 then  -- ok it's an ugly corner case...
              flip_LUT := 0;
              run_mode := probe;
    else
              flip_LUT := flip;
              run_mode := trace;  -- To be changed later
    end if;
    if verb'length > 0 then
      verbose_i := 1;
      if verb="full" then
        verbose_i := 2;
      end if;      
    end if;
    write(xcl_filename, filename);
    GateList := new Gate_Desc_Array(1 to gate_instance_counter);
    indexPtr := new integer_array(0 to gate_instance_counter);
    init_get_vectors; -- load a first exclusion vector into current_vector
  end update_select_gate;

  -- This function is called at the initialisation time of the gate,
  -- after the generics are evaluated and Gate_Census gives the gate_number,
  -- but before the clock starts.
  -- If fast/default mode, the arrays are not initialised and LUT is simply copied.
  -- Otherwise, this allows us to list the gate or initialise and/or tweak its LUT.
  impure function RegisterLUT(L: std_logic_vector; fullpath: string;
       gate_number, lastbit : integer; exclude: std_logic_vector)
  return std_logic_vector is
    variable T : std_logic_vector(L'range) := L;
    variable i : integer;

    procedure exclude_LUT (vector : string) is
      variable i, j : integer;
    begin
--      if vector'length > 0 then  => implicit
      j := 0;
      for i in vector'range loop
        exit when j > lastbit;
        if vector(i) = 'X' then
          T(j             ) := 'X';
          T(j + lastbit+ 1) := 'X';
          GateList(gate_number).histo(j             ):= -1;
          GateList(gate_number).histo(j + lastbit+ 1):= -1;
        end if;
        j := j+1;
      end loop;
      if verbose_i /= 0 then
        report " Excluding inputs of gate #" & integer'image(gate_number) & " -> " & SLV_to_bin(T) & " - " & fullpath;
      end if;
    end exclude_LUT;

  begin
    if run_mode /= fast then
      if gate_select_number < 1 then  -- just a listing
        report " gate#" & integer'image(gate_number)
          & " [bool] : " & fullpath;
      end if;

      GateList(gate_number).gate_kind := Gate_boolean;

      -- save the gate's name in a dynamic area:
      write(GateList(gate_number).gatename, fullpath);

      -- clear the histogram :
      GateList(gate_number).histo := (others => 0);
      GateList(gate_number).inputs := inputs_LUT(lastbit);
      assert GateList(gate_number).inputs >= 0
          report "lastbit's value is invalid (wrong gate input size)"
          severity failure;

      -- register an exclusion :
      if gate_number = xcl_gatenum then
	if current_vector.all'length > 0 then
          exclude_LUT(current_vector.all);
        end if;
        get_another_vector;
      else
        if exclude'length > 0 then
          exclude_LUT(SLV_to_bin(exclude));
        end if;
      end if;

      -- if flip or trace :
      if gate_select_number >= 0 then
        -- init the histogram:
        GateList(gate_number).last := lastbit;

        -- register into the list of updates/display:
        if lastbit > 0 then
          indexPtr(registered_instance_counter) := gate_number;
          registered_instance_counter := registered_instance_counter + 1;
        end if;

        if  gate_select_number = gate_number
         and bit_select_number >= 0
         and bit_select_number <= lastbit then

          assert T(bit_select_number) /= 'X'
             report "Can't flip unreachable bit of gate#" & integer'image(gate_number) & " - " & fullpath
             severity failure;

          if flip_LUT = 1 then
            -- Alter the selected bit
            i := bit_select_number + lastbit+ 1;
            if T(bit_select_number) = '0' then
              T(i) := 'H'; -- invert the high/shadow bits
              T(bit_select_number) := '1';
            else
              T(i) := 'L';
              T(bit_select_number) := '0';
            end if;
            report " Flipping bit " & integer'image(bit_select_number)
                 & " of gate #" & integer'image(gate_number) & " -> " & SLV_to_bin(T) & " - " & fullpath;
          else
            -- transform the selected value into a meta value that will propagate
            if T(bit_select_number) = '0' then
              T(bit_select_number) := 'L';
            else
              T(bit_select_number) := 'H';
            end if;
            report " Tracing bit " & integer'image(bit_select_number)
               & " of gate #" & integer'image(gate_number) & " -> "
               & SLV_to_bin(T) & " - " & fullpath;
          end if;
        end if;
      end if;

      -- Keep a local copy of altered the LUT (for vectgen later)
      GateList(gate_number).LUT(0 to lastbit) := T(0 to lastbit);
    end if;
    return T;
  end RegisterLUT;

  -- same as RegisterLUT but without the LUT stuff.
  impure function RegisterBack(fullpath: string; gate_number : integer) return integer is
  begin
    if run_mode /= fast then
      if gate_select_number < 1 then  -- just a listing
        report " gate#" & integer'image(gate_number) & " [back] : " & fullpath;
      end if;

      -- save the gate's name in a dynamic area:
      write(GateList(gate_number).gatename, fullpath);

      GateList(gate_number).gate_kind := Gate_backwards;
      backwards_instance_counter := backwards_instance_counter;
    end if;
    return gate_number;
  end RegisterBack;


  procedure update_gates_histograms is
    variable i, j, k : integer;
    variable l : big_sint_t;
    variable s, t : std_logic;
  begin
    assert cycle_counter < big_uint_t'high
       report "Simulation reached maximum cycle count"
       severity failure;
    cycle_counter := cycle_counter + 1;
    i := 0;
    while i < registered_instance_counter loop
      j := indexPtr(i);
      k := GateList(j).index;
      l := GateList(j).histo(k);
      assert l /= -1
        report "Gate#" & integer'image(j) & " reached excluded state '" & integer'image(k) & "'"  
        severity failure ;
      GateList(j).histo(k) := l + 1;

      -- count the toggles at the output:
         s := GateList(j).curOut;
      if s /= GateList(j).prevOut then
        GateList(j).changes := GateList(j).changes + 1;
      end if;
      GateList(j).prevOut := s;

      i := i+1;
    end loop;
  end update_gates_histograms;


  procedure display_gates_histograms is
    variable i, j, k, m, n : integer;
    variable c : big_uint_t;
    variable o : big_sint_t;
    variable l : line;
  begin
    report "End of simulation ! " & big_uint_t'image(cycle_counter) & " cycles.";

    c := 0;
    i := 0;
    while i < registered_instance_counter loop
      j := indexPtr(i);
      m := GateList(j).last;
      n := m+m+1;
      for k in 0 to n loop
        o := GateList(j).histo(k);
        if o = -1 then
          write(l, string'("^ "));
        else
          write(l, big_sint_t'image(o) & " ");
        end if;
        if k = m then
          write(l, string'(" |  "));
        end if;
      end loop;

      c := c + GateList(j).changes;

      report " h(" & integer'image(j) & ") = " & L.all
           & " ("  & big_sint_t'image(GateList(j).histo(special))
           & ") [" & big_uint_t'image(GateList(j).changes) & "] - " & GateList(j).gatename.all;
      Deallocate(L);
      i := i+1;
    end loop;

    report "Total state changes: " & big_uint_t'image(c) & " (not counting fanout)";
  end display_gates_histograms;

-----------------------------------------------------------------------------------------
-- Used to probe the netlist :

  -- see also how it registers the input ports in VectGen.vhdl
  impure function register_source(gate_number, input_number: integer; val: std_logic)
   return input_record_type is
    variable this : input_record_type;

    procedure link_net(driver: inout output_record_type) is
      variable count : integer;
    begin
      -- add this gate & port to the linked list of the driver
      this.next_input := driver.first_sink;
      this.next_port  := driver.first_port;
      driver.first_sink := gate_number;
      driver.first_port := input_number;
      count := driver.fanout;
      if count < 0 then
        driver.fanout := 1;
      else
        driver.fanout := count+1;
      end if;
    end procedure;

  begin
    this.source := probe_numbers(val);
    if this.source < 1 then -- is it driven by an input ?
      assert this.source /= integer'low 
         report "Why did I get a signal from an unused slot ? (Is a signal initialised with '0' instead of 'U' ?"
         severity failure;
      if verbose_i > 1 then
        report  "sink#" & integer'image(gate_number) &
                "(" & integer'image(input_number) &
                ") reads " & std_logic'image(val) &
                " from input#" &  integer'image(-(this.source));
      end if;
      link_net(input_records(-(this.source)));
    else  -- it's a gate !
      if verbose_i > 1 then
        report "sink#" & integer'image(gate_number) &
               "(" & integer'image(input_number) &
               ") reads " & std_logic'image(val) &
               " from gate#" &  integer'image(this.source);
      end if;
      link_net(GateList(this.source).out_record); 
    end if;
    return this;
  end register_source;


  procedure register_gate_source(gate_number, input_number: integer; val: std_logic) is
  begin
    -- don't process the gate again if it is already processed:
    if GateList(gate_number).in_record(input_number).source = integer'low then
       GateList(gate_number).in_record(input_number) := register_source(gate_number, input_number, val);
    end if;
  end register_gate_source;


  -- OK this system works but is inefficient...
  impure function probe_return(gate_number : integer) return std_logic
  is
    constant smin : std_logic := '0';
    constant smax : std_logic := '-';
    variable v    : std_logic;
  begin
    for v in smin to smax loop
      if gate_number = probe_numbers(v) then
--        report "gate#" & integer'image(gate_number) & " out: " & std_logic'image(v);
        return v;
      end if;
    end loop;
    return probe_phase;
  end probe_return;


-------------------------------------------------------------------------------------------
-- The function called by each gate :

  impure function GoBackwards (A : std_logic;
      gate_number : integer) return std_logic is
  begin
--    if run_mode = probe then
--        -- match the input
--        if (A /= 'U') and (A /= 'X') then register_gate_source(gate_number, 0, A); end if;
--        return probe_return(gate_number);
--    end if;

    return A;
  end GoBackwards;


  impure function LookItUp1 (A : std_logic; LUT2 : LookupType2;
          gate_number : integer) return std_logic is
    variable index : integer := 0;
    variable R : std_logic := A;
  begin

    case run_mode is
      when fast =>
        --  from LookItUp1_fast:
        if A = '1' then
          return LUT2(1);
        else
          return LUT2(0);
        end if;

      when trace =>
        case A is
          when '0'    => index := 0;
          when '1'    => index := 1;
          when 'L'    => index := 2;
          when 'H'    => index := 3;
          when others => index := SPECIAL;
                GateList(gate_number).index := index;
                return R;
        end case;
        R := LUT2(index);
        GateList(gate_number).index := index;
        GateList(gate_number).curOut := R;
        return R;

      when probe =>
        -- match the input
        if (A /= 'U') and (A /= 'X') then register_gate_source(gate_number, 0, A); end if;
        return probe_return(gate_number);

    end case;
  end LookItUp1;

  impure function LookItUp2 (A, B : std_logic; LUT4 : LookupType4;
          gate_number : integer) return std_logic is
    variable index, meta : integer := 0;
    variable R : std_logic := A;
  begin

    case run_mode is
      when fast =>
        if A = '1' then
          index := 2;
        end if;
        if B = '1' then
          index := index+1;
        end if;
        return LUT4(index);

      when trace =>
        case A is
          when '0'    =>  NULL;
          when '1'    =>  index := 2;
          when 'L'    =>              meta := 4;
          when 'H'    =>  index := 2; meta := 4;
          when others =>  GateList(gate_number).index := SPECIAL; return A;
        end case;

        case B is
          when '0'    =>  NULL;
          when '1'    =>  index := index+1;
          when 'L'    =>                    meta := 4;       
          when 'H'    =>  index := index+1; meta := 4;
          when others =>  GateList(gate_number).index := SPECIAL; return B;
        end case;

        index := index+meta;
        R := LUT4(index);
        GateList(gate_number).index := index;
        GateList(gate_number).curOut := R;
        return R;

      when probe =>
        -- match the input
        if (A /= 'U') and (A /= 'X') then register_gate_source(gate_number, 0, A); end if;
        if (B /= 'U') and (B /= 'X') then register_gate_source(gate_number, 1, B); end if;
        return probe_return(gate_number);

    end case;
  end LookItUp2;

  impure function LookItUp3 (A, B, C : std_logic; LUT8 : LookupType8;
      gate_number : integer) return std_logic is
    variable index, meta : integer := 0;
    variable R : std_logic := A;
  begin

    case run_mode is
      when fast =>
        if A = '1' then
          index := 4;
        end if;

        if B = '1' then
          index := index+2;
        end if;

        if C = '1' then
          index := index+1;
        end if;
        return LUT8(index);

      when trace =>
        case A is
          when '0'    =>  NULL;
          when '1'    =>  index := 4;
          when 'L'    =>              meta := 8;
          when 'H'    =>  index := 4; meta := 8;
          when others =>  GateList(gate_number).index := SPECIAL; return A;
        end case;

        case B is
          when '0'    =>  NULL;
          when '1'    =>  index := index+2;
          when 'L'    =>                    meta := 8;
          when 'H'    =>  index := index+2; meta := 8;
          when others =>  GateList(gate_number).index := SPECIAL; return B;
        end case;

        case C is
          when '0'    =>  NULL;
          when '1'    =>  index := index+1;
          when 'L'    =>                    meta := 8;
          when 'H'    =>  index := index+1; meta := 8;
          when others =>  GateList(gate_number).index := SPECIAL; return C;
        end case;

        index := index+meta;
        R := LUT8(index);
        GateList(gate_number).index := index;
        GateList(gate_number).curOut := R;
        return R;

      when probe =>
        -- match the input
        if (A /= 'U') and (A /= 'X') then register_gate_source(gate_number, 0, A); end if;
        if (B /= 'U') and (B /= 'X') then register_gate_source(gate_number, 1, B); end if;
        if (C /= 'U') and (C /= 'X') then register_gate_source(gate_number, 2, C); end if;
        return probe_return(gate_number);

    end case;
  end LookItUp3;

  impure function LookItUp4 (A, B, C, D : std_logic; LUT16 : LookupType16;
      gate_number : integer) return std_logic is
    variable index, meta : integer := 0;
    variable R : std_logic := A;
  begin

    case run_mode is
      when fast =>
        if A = '1' then
          index := 8;
        end if;

        if B = '1' then
          index := index+4;
        end if;

        if C = '1' then
          index := index+2;
        end if;

        if D = '1' then
          index := index+1;
        end if;
        return LUT16(index);

      when trace =>
        case A is
          when '0'    =>  NULL;
          when '1'    =>  index := 8;
          when 'L'    =>              meta := 16;
          when 'H'    =>  index := 8; meta := 16;
          when others =>  GateList(gate_number).index := SPECIAL; return A;
        end case;

        case B is
          when '0'    =>  NULL;
          when '1'    =>  index := index+4;
          when 'L'    =>                    meta := 16;
          when 'H'    =>  index := index+4; meta := 16;
          when others =>  GateList(gate_number).index := SPECIAL; return B;
        end case;

        case C is
          when '0'    =>  NULL;
          when '1'    =>  index := index+2;
          when 'L'    =>                    meta := 16;
          when 'H'    =>  index := index+2; meta := 16;
          when others =>  GateList(gate_number).index := SPECIAL; return C;
        end case;

        case D is
          when '0'    =>  NULL;
          when '1'    =>  index := index+1;
          when 'L'    =>                    meta := 16;
          when 'H'    =>  index := index+1; meta := 16;
          when others =>  GateList(gate_number).index := SPECIAL; return D;
        end case;

        index := index+meta;
        R := LUT16(index);
        GateList(gate_number).index := index;
        GateList(gate_number).curOut := R;
        return R;

      when probe =>
        -- match the input
        if (A /= 'U') and (A /= 'X') then register_gate_source(gate_number, 0, A); end if;
        if (B /= 'U') and (B /= 'X') then register_gate_source(gate_number, 1, B); end if;
        if (C /= 'U') and (C /= 'X') then register_gate_source(gate_number, 2, C); end if;
        if (D /= 'U') and (D /= 'X') then register_gate_source(gate_number, 3, D); end if;
        return probe_return(gate_number);

    end case;
  end LookItUp4;

-----------------------------------------------------------------------------------------

  -- used by RegisterLUT
  procedure get_another_vector is
    variable char : character;
  begin
    xcl_oldgatenum := xcl_gatenum;
    loop
      if endfile(exclude_file) then
        if verbose_i /= 0 then
          report "end of exclusion input file.";
        end if;
        -- prevent triggering more tests&calls
        xcl_gatenum := -1;
        file_close(exclude_file);
        exit;
      end if;

      line_number := line_number + 1;
      readline(exclude_file, current_vector);

      if current_vector.all'length > 0 then
        if current_vector.all(current_vector.all'left) = '#' then
          if verbose_i /= 0 then
            report "Comment: '" & current_vector.all & "'";
          end if;
        else
          read(current_vector, xcl_gatenum);

          assert xcl_gatenum > xcl_oldgatenum
             report  xcl_filename.all & "(" & integer'image(line_number) & ") : gate#"
             & integer'image(xcl_gatenum) & " index is lower than previous index."
             severity failure;

          assert xcl_gatenum <= gate_instance_counter
            report "The exclude file " & xcl_filename.all & " contains too many vectors"
            severity failure;

          -- skip all the space/separators
          while current_vector.all'length  > 0
           and (current_vector.all(current_vector.all'left) = ' '
             or current_vector.all(current_vector.all'left) = HT ) loop
           read(current_vector, char);
          end loop; -- should be better coded -_-

          -- Little check of the remaining string
          for i in current_vector.all'range loop
            char := current_vector.all(i);
            assert char = '-' or char = 'X'
              report "invalid characters in exclusion string."
              severity failure;
          end loop;

          exit;
        end if;
      end if;
    end loop;
  end get_another_vector;


  procedure init_get_vectors is
  begin
    if xcl_filename.all /= "" then
      file_open(exclude_file, xcl_filename.all, read_mode);
      line_number := 1;
      get_another_vector;
    else
      -- prevent triggering the tests&calls
      if verbose_i /= 0 then
        report "no exclusion input file to read.";
      end if;
      xcl_gatenum := -1;
    end if;
  end init_get_vectors;

end Gates_lib;
