#!/bin/bash

# LibreGates/wrap_netlist.sh
# created sam. sept. 12 09:29:47 CEST 2020
# version mar. sept. 15 05:59:38 CEST 2020 : forked from wrap.sh to become wrap_netlist.sh
# version lun. sept. 21 07:19:31 CEST 2020 : don't call update_select_gate directly.
# version ven. nov.   6 08:07:22  CET 2020 : now you can define $DRIVE_DUT to call a different test driver
# version ven. nov.  20 19:26:00  CET 2020 : tries to propagate the generics
# version sam. nov. 21 10:45:15 CET 2020 : propagates integers and strings generics
#
# Generates a wrapper from the XML output of GHDL

XML_FILE=xml.log

Problem () {
  echo $'\e[91m' $* $'\e[0m' 1>&2
}

ReadTag () {
  IFS=\> read -d \< ENTITY
}

GetAttribute () {      #input list in $1
  LIST=($ATTRIBUTES)   # split string into array
  ATTRIBUTE=""
  for i in $(seq "${#LIST[@]}") ; do  # scan the list
    IFS='"' read -a ATTRIBUTE <<< "${LIST[$i]}"  # split the individual parameter in a sub-string
    if [[ "$ATTRIBUTE" = "$1" ]] ; then   # found the key
      echo -n "${ATTRIBUTE[1]}"           # return the value
    fi
  done
}

[ "$DRIVE_DUT" ] || DRIVE_DUT="Drive_DUT"
PORT_TYPE=""
PORTNAME=""
INOUT=""
LISTMAP=()
INDEX_IN=0
INDEX_OUT=0
CLK_USED="'X'" #this value means unused
GENERIC_MAP=()
GENERIC_DEF=()
GENERIC_NAME=""
GENERIC_TYPE=""
GENERIC_INITVAL=""

RegisterPort () {
  if [[ "$PORT_TYPE" != "" ]] ; then
    if [[ "$PORT_TYPE" == "sl" ]] ||
       [[ "$PORT_TYPE" == "std_logic" ]] ; then
      SIZE=1
    else
      if [[ "$PORT_TYPE" == slv* ]] ; then
        SIZE=${PORT_TYPE#slv}
      else
        SIZE=0
        Problem "Error: $PORT_TYPE not supported"
      fi
    fi

    # apply to the appropriate vector
    case "$INOUT" in
      "in")
        if [[ "$PORTNAME" == "clk" ]] ; then
          if [[ "$PORT_TYPE" == "sl" ]] ||
             [[ "$PORT_TYPE" == "std_logic" ]] ; then
            LISTMAP+=('clk => VectClk')
            CLK_USED="'U'"
          else
            Problem "Invalid clock signal type : $PORT_TYPE"
          fi
        else
          RANGE=$INDEX_IN
          if [ "$SIZE" -gt 1 ]; then
            RANGE="$(( INDEX_IN - 1 + SIZE )) downto $INDEX_IN"
          fi;
          LISTMAP+=("$PORTNAME => VectIn($RANGE)")
          INDEX_IN=$(( INDEX_IN + SIZE ))
        fi
        ;;
      "out")
        RANGE=$INDEX_OUT
        if [ "$SIZE" -gt 1 ]; then
          RANGE="$(( INDEX_OUT - 1 + SIZE )) downto $INDEX_OUT"
        fi;
        LISTMAP+=("$PORTNAME => VectOut($RANGE)")
        INDEX_OUT=$(( INDEX_OUT + SIZE ))
        ;;
      *) Problem "Invalid port direction for signal $PORTNAME"
    esac
  fi
}

RegisterGeneric () {
  [ "$GENERIC_NAME" ] && {
    GENERIC_MAP+=("$GENERIC_NAME => $GENERIC_NAME")

    GEN="$GENERIC_NAME : $GENERIC_TYPE"
    [ "$GENERIC_INITVAL" ] &&
      GEN=$GEN" := "$DELIMITER$GENERIC_INITVAL$DELIMITER
    GENERIC_DEF+=("$GEN,")

    SEEK_NAME="default_value"
    SEEK_ATTR="value="
    GENERICS="1"
    GENERIC_NAME=""
    GENERIC_TYPE=""
    GENERIC_INITVAL=""
    DELIMITER=''
  }
}

#æssume <generic_chain> comes before <port_chain>
Seek_Tag () {
  while ReadTag; do
    TAG_NAME=${ENTITY%% *}
    if [[ "$TAG_NAME" == "$1" ]] ; then
      return
    fi
  done
}

Seek_port_chain () {


}

# Get the top level entity:
ghdl -a $* &&
TOP_ENTITY=$(ghdl --find-top $*) &&
ghdl --file-to-xml $* > $XML_FILE &&  # transform the VHDL into XML

# parse the stream and look for the ports:
while ReadTag; do
  TAG_NAME=${ENTITY%% *}
  ATTRIBUTES=${ENTITY#* }

  # executed in sequence :
  if [[ "$TAG_NAME" == "el" ]] ; then
    RegisterPort
    INOUT=$(GetAttribute 'mode=')
    PORTNAME=$(GetAttribute 'identifier=')
  else
    if [[ "$TAG_NAME" == "subtype_indication" ]] ; then
      PORT_TYPE=$(GetAttribute 'identifier=')
    fi
  fi
done <<< $( # don't use a pipe because it would create a subshell
            # and the collected variables would be lost...
  # Now parse the AST for the ports list and store the text in a variable :
  # keep only the first "port_chain" block which should be the top-level interface
  sed  '/ <\/port_chain>$/,$d' $XML_FILE | # drop everything after the first end of block
  sed -n '/ <port_chain>$/,$p' ) && # drop everything before the first start of block

RegisterPort && #flush the buffer

# Check and duplicate the generics
GENERICS="" &&

grep -q '<generic_chain>' $XML_FILE  && {
  SEEK_NAME="default_value"
  SEEK_ATTR="value="
  DELIMITER=''
  while ReadTag; do
    TAG_NAME=${ENTITY%% *}
    ATTRIBUTES=${ENTITY#* }
    if [[ "$TAG_NAME" == "el" ]] ; then
      RegisterGeneric
      GENERIC_NAME=$(GetAttribute 'identifier=')
    else
      if [[ "$TAG_NAME" == "subtype_indication" ]] ; then
        GENERIC_TYPE=$(GetAttribute 'identifier=')
      else
        if [[ "$TAG_NAME" == "$SEEK_NAME" ]] ; then
          GENERIC_INITVAL=$(GetAttribute "$SEEK_ATTR")
          # some types require more processing,
          # the contents is in another tag when value= is missing
          if [[ "$GENERIC_INITVAL" == "" ]] ; then
            if [[ "$GENERIC_TYPE" == "string" ]] ; then
              SEEK_NAME="string8_id"
              SEEK_ATTR="content="
              DELIMITER='"'
            fi
            # not enough to make a case yet.
          fi
        fi
      fi
    fi
  done <<< $(
    sed  '/ <\/generic_chain>$/,$d' $XML_FILE | # drop everything after the first end of block
    sed -n '/ <generic_chain>$/,$p' | # drop everything before the first start of block
    sed 's/=" /="/g'  # why is there a leading space inside an attribute ???
  ) &&
  RegisterGeneric # flush
}

if   [ "$INDEX_IN"  -gt 0 ]; then
  if [ "$INDEX_OUT" -gt 0 ]; then

cat <<ENDOFHEREDOC
-- Wrapper.vhdl
-- Generated on $(date +%Y%m%d-%R)
-- Do not modify

Library ieee;
    use ieee.std_logic_1164.all;
Library work;
    use work.all;
Library LibreGates;
    use LibreGates.all;
    use LibreGates.Gates_lib.all;
    use LibreGates.Netlist_lib.all;

entity Wrapper is
  generic (
ENDOFHEREDOC

if [[ "$GENERICS" ]] ; then
  for i  in "${!GENERIC_MAP[@]}"; do
    echo "    ${GENERIC_DEF[$i]}"
  done
fi

cat <<ENDOFHEREDOC
    WrapperWidthIn : integer := $INDEX_IN;
    WrapperWidthOut: integer := $INDEX_OUT;
    listgates: integer := 1;  -- set to 0 to display the raw list of gates
    main_delay : time := 1 sec;
    filename : string := ""; -- file of exclude vectors
    verbose : string := ""
  );
end Wrapper;

architecture Wrap of Wrapper is
  signal VectIn : std_logic_vector(WrapperWidthIn -1 downto 0);
  signal VectOut: std_logic_vector(WrapperWidthOut-1 downto 0);
  signal VectClk: std_logic := $CLK_USED;

begin

  -- force the registration of the gates.
  -- must be placed at the very beginning
  rg: entity register_generics generic map(
         gate_select_number => listgates, -- show gate listing when < 1
         bit_select_number => -1, -- no alteration
         flip_LUT => 2,           -- probe mode
         filename => filename,    -- eventual exclusions
         verbose => verbose)
  port map(clock => '0');   -- the updates are directly driven by $DRIVE_DUT
  -- note : the corresponding procedure is not called
  -- at the right time if it is called directly,
  -- the entity ensures the correct calling order.

  $DRIVE_DUT(VectIn, VectOut, VectClk, main_delay);

  -- Finally we "wire" the unit to the ports:
  tb: entity $TOP_ENTITY
ENDOFHEREDOC

if [[ "$GENERICS" ]] ; then
  echo "    generic map("
  COMA=""
  for i in "${!GENERIC_MAP[@]}"; do
    [[ "$COMA" ]] && echo $COMA; COMA=","
    echo -n "      ${GENERIC_MAP[$i]}"
  done
  echo ");"
fi

echo "    port map("
COMA=""
for i in "${!LISTMAP[@]}"; do
  [[ "$COMA" ]] && echo $COMA; COMA=","
  echo -n "      ${LISTMAP[$i]}"
done

echo ");"
echo "end Wrap;"
  else
    Problem "Error: no output found"
  fi
else
  Problem "Error: no input found"
fi

#rm -f $XML_FILE
